from typing import Optional, List, Dict, Any
import httpx
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from pydantic import BaseModel, Field
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async


class EvaluatorAPIDTO(BaseModel):
    id: str = Field(alias="id")
    alias: str = Field(alias="alias")
    version: str = Field(alias="version")
    label: Optional[str] = Field(alias="label")
    description: Optional[str] = Field(alias="description")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")
    specs: Dict = Field(alias="specs")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateEvaluatorDTO(BaseModel):
    label: Optional[str] = Field(alias="label")
    alias: str = Field(alias="alias")
    version: str = Field(alias="version")
    description: Optional[str] = Field(alias="description")
    specs: Dict = Field(alias="specs")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateEvaluatorDTO(BaseModel):
    label: Optional[str] = Field(alias="label")
    description: Optional[str] = Field(alias="description")
    specs: Dict = Field(alias="specs")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class Instance(BaseModel):
    reference_id: str = Field(alias="referenceId")
    reference_date: str = Field(alias="referenceDate")
    data: Dict[str, Any] = Field(alias="data", default={})


class Entity(BaseModel):
    entity_id: str = Field(alias="entityId")
    role: Optional[str] = Field(alias="role", default=None)
    character: Optional[str] = Field(alias="character", default=None)
    data: Dict[str, Any] = Field(alias="data", default={})


class EvaluatorInput(BaseModel):
    instance: Instance = Field(alias="instance")
    entities: List[Entity] = Field(alias="entities")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BusinessRuleOutput(BaseModel):
    id: Optional[str] = Field(alias="id", default=None)
    order: int = Field(alias="order")
    code: str = Field(alias="code")
    label: str = Field(alias="label")
    value: str = Field(alias="value")
    alert_level: Optional[int] = Field(alias="alertLevel", default=None)
    # hit can be None if the rule cannot be evaluated due to missing fields
    hit: Optional[bool] = Field(alias="hit", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ScoreOutput(BaseModel):
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    value: float = Field(alias="value")
    max_value: Optional[float] = Field(alias="maxValue", default=None)

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ScoreCardRuleOutput(BaseModel):
    field: str = Field(alias="field")
    order: int = Field(alias="order")
    value: Any = Field(alias="value")
    bucket: int = Field(alias="bucket")
    points: int = Field(alias="points")
    max_points: int = Field(alias="maxPoints")
    label: str = Field(alias="label")
    bucket_label: str = Field(alias="bucketLabel")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class MetricOutput(BaseModel):
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    value: Any = Field(alias="value")
    metadata: Optional[Dict] = Field(alias="metadata", default=None)


class EvaluatorOutput(BaseModel):
    score: Optional[ScoreOutput] = Field(alias="score", default=None)
    scorecard: List[ScoreCardRuleOutput] = Field(alias="scorecard", default=[])
    metrics: List[MetricOutput] = Field(alias="metrics", default=[])
    rules: List[BusinessRuleOutput] = Field(alias="rules", default=[])
    decision: str = Field(alias="decision")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class EvaluatorOutputError(BaseModel):
    detail: str = Field(alias="detail")
    traceback: List[str] = Field(alias="traceback")

    class Config:
        allow_population_by_field_name = True
        allow_population_by_alias = True


class EvaluatorSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "evaluators", header_builder, renew_token, EvaluatorAPIDTO.parse_obj(data))

    @retry_on_401
    def evaluate(self, evaluator_input: EvaluatorInput) -> EvaluatorOutput | EvaluatorOutputError:
        with httpx.Client() as client:
            response = client.post(
                f"{self.base_url}/{self.resource}/{self.data.id}/evaluate",
                headers=self._header_builder(),
                json=evaluator_input.dict(by_alias=True),
                timeout=300
            )
            raise_for_status_improved(response)
            if "traceback" in response.json():
                return EvaluatorOutputError.parse_obj(response.json())
            else:
                return EvaluatorOutput.parse_obj(response.json())


class EvaluatorAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "evaluators", header_builder, renew_token, EvaluatorAPIDTO.parse_obj(data))

    @retry_on_401_async
    async def evaluate(self, evaluator_input: EvaluatorInput) -> EvaluatorOutput | EvaluatorOutputError:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/{self.resource}/{self.data.id}/evaluate",
                headers=self._header_builder(),
                json=evaluator_input.dict(by_alias=True),
                timeout=300
            )
            raise_for_status_improved(response)
            if "traceback" in response.json():
                return EvaluatorOutputError.parse_obj(response.json())
            else:
                return EvaluatorOutput.parse_obj(response.json())


class EvaluatorSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=EvaluatorSync,
                         retrieve_data_model=EvaluatorAPIDTO,
                         create_data_model=CreateEvaluatorDTO,
                         update_data_model=UpdateEvaluatorDTO,
                         resource="evaluators")

    def renew_token(self):
        self.altscore_client.renew_token()

    @retry_on_401
    def evaluate(
            self, evaluator_input: Dict, evaluator_id: Optional[str] = None,
            evaluator_alias: Optional[str] = None, evaluator_version: Optional[str] = None
    ) -> EvaluatorOutput | EvaluatorOutputError:

        if evaluator_id is not None:
            url = f"{self.altscore_client._borrower_central_base_url}/v1/evaluators/{evaluator_id}/evaluate"
        elif evaluator_alias is not None and evaluator_version is not None:
            url = f"{self.altscore_client._borrower_central_base_url}/v1/evaluators/{evaluator_alias}/{evaluator_version}/evaluate"
        else:
            raise ValueError("either evaluator_id or evaluator_alias and evaluator_version must be provided")

        with httpx.Client() as client:
            response = client.post(
                url,
                headers=self.build_headers(),
                json=EvaluatorInput.parse_obj(evaluator_input).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            if "traceback" in response.json():
                return EvaluatorOutputError.parse_obj(response.json())
            else:
                return EvaluatorOutput.parse_obj(response.json())


class EvaluatorAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=EvaluatorAsync,
                         retrieve_data_model=EvaluatorAPIDTO,
                         create_data_model=CreateEvaluatorDTO,
                         update_data_model=UpdateEvaluatorDTO,
                         resource="evaluators")

    def renew_token(self):
        self.altscore_client.renew_token()

    @retry_on_401_async
    async def evaluate(
            self, evaluator_input: Dict, evaluator_id: Optional[str] = None,
            evaluator_alias: Optional[str] = None, evaluator_version: Optional[str] = None
    ) -> EvaluatorOutput | EvaluatorOutputError:

        if evaluator_id is not None:
            url = f"{self.altscore_client._borrower_central_base_url}/v1/evaluators/{evaluator_id}/evaluate"
        elif evaluator_alias is not None and evaluator_version is not None:
            url = f"{self.altscore_client._borrower_central_base_url}/v1/evaluators/{evaluator_alias}/{evaluator_version}/evaluate"
        else:
            raise ValueError("either evaluator_id or evaluator_alias and evaluator_version must be provided")

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.build_headers(),
                json=EvaluatorInput.parse_obj(evaluator_input).dict(by_alias=True),
                timeout=120
            )
            raise_for_status_improved(response)
            if "traceback" in response.json():
                return EvaluatorOutputError.parse_obj(response.json())
            else:
                return EvaluatorOutput.parse_obj(response.json())



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
import datetime as dt

class HistoricValue(BaseModel):
    reference_id: str = Field(alias="referenceId")  # this is the id an identifier for the source of the value
    value: Any = Field(alias="value")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BorrowerFieldAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    value: Any = Field(alias="value")
    data_type: str = Field(alias="dataType")
    history: List[HistoricValue] = Field(alias="history")
    tags: List[str] = Field(alias="tags", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateBorrowerFieldDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    form_id: Optional[str] = Field(alias="formId", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)
    key: str = Field(alias="key")
    value: Any = Field(alias="value")
    data_type: Optional[str] = Field(alias="dataType", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    updated_at: Optional[dt.datetime] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    def dict(self, *args, **kwargs):
        base_dict = super().dict(*args, **kwargs)
        date_key = 'updatedAt' if kwargs.get("by_alias") else 'updated_at'
        base_dict[date_key] = self.updated_at.isoformat() if self.updated_at else None
        return base_dict


class UpdateBorrowerFieldDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    form_id: Optional[str] = Field(alias="formId", default=None)
    reference_id: Optional[str] = Field(alias="referenceId", default=None)
    value: Optional[str] = Field(alias="value")
    data_type: Optional[str] = Field(alias="dataType", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    updated_at: Optional[dt.datetime] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    def dict(self, *args, **kwargs):
        base_dict = super().dict(*args, **kwargs)
        date_key = 'updatedAt' if kwargs.get("by_alias") else 'updated_at'
        base_dict[date_key] = self.updated_at.isoformat() if self.updated_at else None
        return base_dict


class BorrowerFieldSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "borrower-fields", header_builder, renew_token, BorrowerFieldAPIDTO.parse_obj(data))


class BorrowerFieldAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "borrower-fields", header_builder, renew_token, BorrowerFieldAPIDTO.parse_obj(data))


class BorrowerFieldsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=BorrowerFieldSync,
                         retrieve_data_model=BorrowerFieldAPIDTO,
                         create_data_model=CreateBorrowerFieldDTO,
                         update_data_model=UpdateBorrowerFieldDTO,
                         resource="borrower-fields")

    @retry_on_401
    def find_by_key(self, key: str, borrower_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            fields_found_req = client.get(
                f"/v1/borrower-fields",
                params={
                    "key": key,
                    "borrower-id": borrower_id,
                    "per-page": 1,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(fields_found_req)
            fields_found_data = fields_found_req.json()
            if len(fields_found_data) == 0:
                return None
            else:
                return self.retrieve(fields_found_data[0]["id"])

    def count_distinct_values(self, key: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            unique_values_req = client.get(
                f"/v1/borrower-fields/queries/count-distinct-values",
                params={
                    "key": key
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(unique_values_req)
            return unique_values_req.json()

    def bulk_update_field_values(self, key: str, current_value: str, target_value: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                f"/v1/borrower-fields/commands/bulk-update-values",
                json={
                    "key": key,
                    "currentValue": current_value,
                    "targetValue": target_value
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return

    @retry_on_401
    def bulk_update_by_borrower_ids(self, borrower_ids: List[str], key: str, new_value: Any, reference_id: Optional[str] = None):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            payload = {
                "borrowerIds": borrower_ids,
                "key": key,
                "newValue": new_value
            }
            if reference_id is not None:
                payload["referenceId"] = reference_id
                
            response = client.post(
                f"/v1/borrower-fields/commands/bulk-update-by-borrower-ids",
                json=payload,
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return

class BorrowerFieldsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=BorrowerFieldAsync,
                         retrieve_data_model=BorrowerFieldAPIDTO,
                         create_data_model=CreateBorrowerFieldDTO,
                         update_data_model=UpdateBorrowerFieldDTO,
                         resource="borrower-fields")

    @retry_on_401_async
    async def find_by_key(self, key: str, borrower_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            fields_found_req = await client.get(
                f"/v1/borrower-fields",
                params={
                    "key": key,
                    "borrower-id": borrower_id,
                    "per-page": 1,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(fields_found_req)
            fields_found_data = fields_found_req.json()
            if len(fields_found_data) == 0:
                return None
            else:
                return self.retrieve(fields_found_data[0]["id"])


    async def count_distinct_values(self, key: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            unique_values_req = await client.get(
                f"/v1/borrower-fields/queries/count-distinct-values",
                params={
                    "key": key
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(unique_values_req)
            return unique_values_req.json()


    async def bulk_update_field_values(self, key: str, current_value: str, target_value: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/borrower-fields/commands/bulk-update-values",
                json={
                    "key": key,
                    "currentValue": current_value,
                    "targetValue": target_value
                },
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return

    @retry_on_401_async
    async def bulk_update_by_borrower_ids(self, borrower_ids: List[str], key: str, new_value: Any, reference_id: Optional[str] = None):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            payload = {
                "borrowerIds": borrower_ids,
                "key": key,
                "newValue": new_value
            }
            if reference_id is not None:
                payload["referenceId"] = reference_id
                
            response = await client.post(
                f"/v1/borrower-fields/commands/bulk-update-by-borrower-ids",
                json=payload,
                headers=self.build_headers(),
                timeout=120
            )
            raise_for_status_improved(response)
            return



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any, Union
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from altscore.borrower_central.model.deal_steps import DealStepDTO


# DTO for current step in a deal
class StepDataDTO(BaseModel):
    """Data transfer object for a step in a deal"""
    step_id: str = Field(alias="stepId")
    order: int = Field(alias="order")
    key: str = Field(alias="key")
    label: Optional[str] = Field(alias="label", default=None)
    created_at: str = Field(alias="createdAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# DTO for deals
class DealDTO(BaseModel):
    """Data transfer object for deals"""
    id: str = Field(alias="id")
    external_id: Optional[str] = Field(alias="externalId", default=None)
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    current_step: Optional[StepDataDTO] = Field(alias="currentStep", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateDealRequest(BaseModel):
    """Model for creating a new deal"""
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    external_id: Optional[str] = Field(alias="externalId", default=None)
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateDealRequest(BaseModel):
    """Model for updating a deal"""
    name: Optional[str] = Field(alias="name", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    status: Optional[str] = Field(alias="status", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ExternalIdRequest(BaseModel):
    """Model for setting an external ID"""
    external_id: str = Field(alias="externalId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class PutCurrentStepRequest(BaseModel):
    """Model for setting the current step of a deal"""
    key: str = Field(alias="key")
    comment: Optional[str] = Field(alias="comment", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# Resource classes for deals
class DealSync(GenericSyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "deals", header_builder, renew_token, DealDTO.parse_obj(data))

    @retry_on_401
    def get_current_step(self):
        """
        Get the current step for a deal
        
        Returns:
            DealStepDTO: The current step
        """
        from altscore.borrower_central.model.deal_steps import DealStepSync

        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                f"{self.base_url}/v1/deals/{self.data.id}/steps/current",
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            return DealStepSync(
                base_url=self.base_url,
                header_builder=self._header_builder,
                renew_token=self.renew_token,
                data=response.json()
            )

    @retry_on_401
    def set_current_step(self, key: str, comment: Optional[str] = None):
        """
        Set the current step for a deal
        
        Args:
            key: The key of the step to set as current
            comment: Optional comment for the step change
            
        Returns:
            None
        """
        with httpx.Client(base_url=self.base_url) as client:
            request_data = PutCurrentStepRequest(key=key, comment=comment)
            response = client.put(
                f"{self.base_url}/v1/deals/{self.data.id}/steps/current",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self._header_builder()
            )
            raise_for_status_improved(response)

    @retry_on_401
    def get_steps(self):
        """
        Get all steps for this deal
        
        Returns:
            List[DealStepDTO]: List of all steps for this deal
        """
        from altscore.borrower_central.model.deal_steps import DealStepDTO

        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                f"{self.base_url}/v1/deals/{self.data.id}/steps",
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            return [DealStepDTO.parse_obj(data) for data in response.json()]


class DealAsync(GenericAsyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "deals", header_builder, renew_token, DealDTO.parse_obj(data))

    @retry_on_401_async
    async def get_current_step(self):
        """
        Get the current step for a deal
        
        Returns:
            DealStepDTO: The current step
        """
        from altscore.borrower_central.model.deal_steps import DealStepAsync

        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                f"{self.base_url}/v1/deals/{self.data.id}/steps/current",
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            return DealStepAsync(
                base_url=self.base_url,
                header_builder=self._header_builder,
                renew_token=self.renew_token,
                data=response.json()
            )

    @retry_on_401_async
    async def set_current_step(self, key: str, comment: Optional[str] = None):
        """
        Set the current step for a deal
        
        Args:
            key: The key of the step to set as current
            comment: Optional comment for the step change
            
        Returns:
            None
        """
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            request_data = PutCurrentStepRequest(key=key, comment=comment)
            response = await client.put(
                f"{self.base_url}/v1/deals/{self.data.id}/steps/current",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self._header_builder()
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def get_steps(self):
        """
        Get all steps for this deal
        
        Returns:
            List[DealStepDTO]: List of all steps for this deal
        """
        from altscore.borrower_central.model.deal_steps import DealStepDTO

        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                f"{self.base_url}/v1/deals/{self.data.id}/steps",
                headers=self._header_builder()
            )
            raise_for_status_improved(response)
            return [DealStepDTO.parse_obj(data) for data in response.json()]


# Module for deals - synchronous
class DealsSyncModule(GenericSyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=DealSync,
                         retrieve_data_model=DealDTO,
                         create_data_model=CreateDealRequest,
                         update_data_model=UpdateDealRequest,
                         resource="deals")

    @retry_on_401
    def set_external_id(self, deal_id: str, external_id: str):
        """
        Set an external ID for a deal
        
        Args:
            deal_id: The ID of the deal
            external_id: The external ID to set
            
        Returns:
            None
        """
        request_data = ExternalIdRequest(externalId=external_id)

        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.put(
                f"/v1/deals/{deal_id}/external-id",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401
    def query_by_borrower_id(self, borrower_id: str, page: int = 1, per_page: int = 10):
        """
        Find deals by borrower ID
        
        Args:
            borrower_id: The ID of the borrower to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with deals and pagination info
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/deals",
                params={
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401
    def query_by_status(self, status: str, page: int = 1, per_page: int = 10):
        """
        Find deals by status
        
        Args:
            status: The status to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with deals and pagination info
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                f"/v1/deals",
                params={
                    "status": status,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()


# Module for deals - asynchronous
class DealsAsyncModule(GenericAsyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=DealAsync,
                         retrieve_data_model=DealDTO,
                         create_data_model=CreateDealRequest,
                         update_data_model=UpdateDealRequest,
                         resource="deals")

    @retry_on_401_async
    async def set_external_id(self, deal_id: str, external_id: str):
        """
        Set an external ID for a deal
        
        Args:
            deal_id: The ID of the deal
            external_id: The external ID to set
            
        Returns:
            None
        """
        request_data = ExternalIdRequest(externalId=external_id)

        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.put(
                f"/v1/deals/{deal_id}/external-id",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self.build_headers(),
                timeout=120,
            )
            await raise_for_status_improved(response)
            return None

    @retry_on_401_async
    async def query_by_borrower_id(self, borrower_id: str, page: int = 1, per_page: int = 10):
        """
        Find deals by borrower ID
        
        Args:
            borrower_id: The ID of the borrower to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with deals and pagination info
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/deals",
                params={
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            await raise_for_status_improved(response)
            return response.json()

    @retry_on_401_async
    async def query_by_status(self, status: str, page: int = 1, per_page: int = 10):
        """
        Find deals by status
        
        Args:
            status: The status to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with deals and pagination info
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                f"/v1/deals",
                params={
                    "status": status,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            await raise_for_status_improved(response)
            return response.json()



from typing import Optional


def build_headers(module, partner_id: Optional[str] = None):
    headers = {}
    if isinstance(module.altscore_client.api_key, str):
        headers["X-API-KEY"] = module.altscore_client.api_key
    elif isinstance(module.altscore_client.user_token, str):
        user_token = module.altscore_client.user_token.replace("Bearer ", "")
        headers["Authorization"] = f"Bearer {user_token}"
    # This is important to avoid infinite recursion
    if partner_id == "init":
        return headers
    elif isinstance(partner_id, str):
        headers["X-PARTNER-ID"] = partner_id
    # if it reaches here when querying the partner it will trigger an infinite recursion
    elif isinstance(module.altscore_client.partner_id, str):
        headers["X-PARTNER-ID"] = module.altscore_client.partner_id
    return headers



from pydantic import BaseModel, Field
from typing import Optional
import datetime as dt


class CurrentDecisionInExecution(BaseModel):
    decision_id: str = Field(alias="decisionId")
    key: str = Field(alias="key")
    decision_type: str = Field(alias="decisionType")
    principal_id: str = Field(alias="principalId")
    updated_at: str = Field(alias="updatedAt")


class PostDecisionToExecution(BaseModel):
    key: str = Field(alias="key")
    decision_type: str = Field(alias="decisionType")
    principal_id: Optional[str] = Field(alias="principalId")
    updated_at: Optional[dt.datetime] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True



from pydantic import BaseModel, Field


class SourceConfig(BaseModel):
    source_id: str = Field(alias="sourceId")
    version: str = Field(alias="version")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True



from pydantic import BaseModel, Field
from typing import Optional
from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule


class AmountInfo(BaseModel):
    amount: str = Field(alias="amount")
    currency: str = Field(alias="currency")
    display: Optional[str] = Field(None, alias="display")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class ClientInfo(BaseModel):
    client_id: str = Field(alias="clientId")
    external_id: str = Field(alias="externalId")
    legal_name: str = Field(alias="legalName")
    tax_id: str = Field(alias="taxId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PartnerInfo(BaseModel):
    name: str = Field(alias="name")
    partner_id: str = Field(alias="partnerId")
    tax_id: str = Field(alias="taxId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PayerInfo(BaseModel):
    account_number: str = Field(alias="accountNumber")
    institution: int = Field(alias="institution")
    institution_name: str = Field(alias="institutionName")
    name: str = Field(alias="name")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PaymentOrderAPIDTO(BaseModel):
    id: str = Field(alias="payOrderId")
    amount: AmountInfo = Field(alias="amount")
    amount_transferred: AmountInfo = Field(alias="amountTransferred")
    client: ClientInfo = Field(alias="client")
    created_at: str = Field(alias="createdAt")
    debt_id: str = Field(alias="debtId")
    disbursement_date: str = Field(alias="disbursementDate")
    gateway: str = Field(alias="gateway")
    notes: Optional[str] = Field(alias="notes")
    partner: PartnerInfo = Field(alias="partner")
    payer: Optional[PayerInfo] = Field(alias="payer", default=None)
    payment_date: Optional[str] = Field(alias="paymentDate", default=None)
    reference_id: str = Field(alias="referenceId")
    reference_number: int = Field(alias="referenceNumber")
    status: str = Field(alias="status")
    tracking_key: str = Field(alias="trackingKey")
    updated_at: str = Field(alias="updatedAt")
    version: int = Field(alias="version")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PaymentOrdersBase:

    @staticmethod
    def _payment_orders():
        return f"/v1/payment-orders"


class PaymentOrdersAsync(PaymentOrdersBase):
    data: PaymentOrderAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: PaymentOrderAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class PaymentOrdersSync(PaymentOrdersBase):
    data: PaymentOrderAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: PaymentOrderAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class PaymentOrdersAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=PaymentOrdersAsync,
            retrieve_data_model=PaymentOrderAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="payment-orders"
        )


class PaymentOrdersSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=PaymentOrdersSync,
            retrieve_data_model=PaymentOrderAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="payment-orders"
        )



from pydantic import BaseModel, Field
from typing import Optional, List, Any, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class Money(BaseModel):
    """
    Simple monetary object with string amount for precision
    """
    amount: str = Field(alias="amount", description="Amount as string to preserve precision")
    currency: str = Field(alias="currency", description="ISO-4217 currency code")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BankAccountAPIDTO(BaseModel):
    """
    Bank account information for tenant-level financial operations
    """
    id: str = Field(alias="id", description="Unique bank account identifier")
    account_number: str = Field(alias="accountNumber", description="Bank account number")
    account_name: str = Field(alias="accountName", description="Display name for the account")
    bank_name: str = Field(alias="bankName", description="Financial institution name")
    bank_code: Optional[str] = Field(alias="bankCode", default=None, description="Bank routing or swift code")
    currency: str = Field(alias="currency", description="Account's primary currency")
    account_type: Optional[str] = Field(alias="accountType", default=None, description="Account type: checking, savings, etc.")
    is_active: bool = Field(alias="isActive", default=True, description="Whether the account is active")
    account_status: Optional[str] = Field(alias="accountStatus", default=None, description="Current account status")
    tags: List[str] = Field(alias="tags", default_factory=list, description="Classification tags")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None, description="Additional custom data")
    created_at: str = Field(alias="createdAt", description="Account creation timestamp")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None, description="Last update timestamp")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateBankAccountDTO(BaseModel):
    """
    Request payload for creating a new bank account
    """
    account_number: str = Field(alias="accountNumber")
    account_name: str = Field(alias="accountName")
    bank_name: str = Field(alias="bankName")
    bank_code: Optional[str] = Field(alias="bankCode", default=None)
    currency: str = Field(alias="currency")
    account_type: Optional[str] = Field(alias="accountType", default=None)
    is_active: bool = Field(alias="isActive", default=True)
    account_status: Optional[str] = Field(alias="accountStatus", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateBankAccountDTO(BaseModel):
    """
    Request payload for updating an existing bank account
    """
    account_name: Optional[str] = Field(alias="accountName", default=None)
    bank_name: Optional[str] = Field(alias="bankName", default=None)
    bank_code: Optional[str] = Field(alias="bankCode", default=None)
    account_type: Optional[str] = Field(alias="accountType", default=None)
    is_active: Optional[bool] = Field(alias="isActive", default=None)
    account_status: Optional[str] = Field(alias="accountStatus", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BankAccountSync(GenericSyncResource):
    """
    Synchronous resource for bank account operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/accounts", header_builder,
                         renew_token, BankAccountAPIDTO.parse_obj(data))


class BankAccountAsync(GenericAsyncResource):
    """
    Asynchronous resource for bank account operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/accounts", header_builder,
                         renew_token, BankAccountAPIDTO.parse_obj(data))


class BankAccountsSyncModule(GenericSyncModule):
    """
    Synchronous module for bank account management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=BankAccountSync,
                         retrieve_data_model=BankAccountAPIDTO,
                         create_data_model=CreateBankAccountDTO,
                         update_data_model=UpdateBankAccountDTO,
                         resource="banking/accounts")

class BankAccountsAsyncModule(GenericAsyncModule):
    """
    Asynchronous module for bank account management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=BankAccountAsync,
                         retrieve_data_model=BankAccountAPIDTO,
                         create_data_model=CreateBankAccountDTO,
                         update_data_model=UpdateBankAccountDTO,
                         resource="banking/accounts")


from json import loads
from functools import wraps
from httpx import HTTPStatusError
from loguru import logger


def retry_on_401(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except HTTPStatusError as e:
            if e.response.status_code == 401:
                logger.info("Token expired, renewing and retrying")
                args[0].renew_token()  # Assuming the first argument is self
                logger.info("Token renewed, retrying")
                return f(*args, **kwargs)
            else:
                raise

    return wrapper


def retry_on_401_async(f):
    async def wrapper(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except HTTPStatusError as e:
            if e.response.status_code == 401:
                logger.info("Token expired, renewing and retrying")
                args[0].renew_token()  # Assuming the first argument is self
                logger.info("Token renewed, retrying")
                return await f(*args, **kwargs)
            else:
                raise

    return wrapper


def raise_for_status_improved(httpx_request) -> None:
    request = httpx_request._request
    if request is None:
        raise RuntimeError(
            "something went wrong, please report this issue"
        )

    if httpx_request.is_success:
        return

    # Try to parse JSON response for more details
    try:
        parsed_text = loads(httpx_request.text)
    except:
        parsed_text = httpx_request.text[:200]  # First 200 characters if not JSON

    message = (
        "{error_type} '{0.status_code} {0.reason_phrase}' for url '{0.url}'\n"
    )

    status_class = httpx_request.status_code // 100
    error_types = {
        1: "Informational response",
        3: "Redirect response",
        4: "Client error",
        5: "Server error",
    }
    error_type = error_types.get(status_class, "Invalid status code")
    message = message.format(httpx_request, error_type=error_type) + "Response: " + str(parsed_text) + "\n"
    raise HTTPStatusError(message, request=request, response=httpx_request)



def build_headers(module):
    if isinstance(module.altscore_client.api_key, str):
        return {"API-KEY": module.altscore_client.api_key}
    elif isinstance(module.altscore_client.form_token, str):
        form_token = module.altscore_client.form_token.replace("Bearer ", "")
        return {"Authorization": f"Bearer {form_token}"}
    elif isinstance(module.altscore_client.user_token, str):
        user_token = module.altscore_client.user_token.replace("Bearer ", "")
        return {"Authorization": f"Bearer {user_token}"}
    return {}



from typing import Optional
from pydantic import BaseModel, Field
import httpx

from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.cms.model.generics import GenericAsyncModule, GenericSyncModule

class AmountInfo(BaseModel):
    amount: str = Field(alias="amount")
    currency: str = Field(alias="currency")
    display: Optional[str] = Field(None, alias="display")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class ClientInfo(BaseModel):
    client_id: str = Field(alias="clientId")
    external_id: str = Field(alias="externalId")
    legal_name: str = Field(alias="legalName")
    tax_id: str = Field(alias="taxId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PartnerInfo(BaseModel):
    name: str = Field(alias="name")
    partner_id: str = Field(alias="partnerId")
    tax_id: str = Field(alias="taxId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PayerInfo(BaseModel):
    account_number: str = Field(alias="accountNumber")
    institution: int = Field(alias="institution")
    institution_name: str = Field(alias="institutionName")
    name: str = Field(alias="name")
    tax_id: Optional[str] = Field(alias="rfcCurp", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DisbursementAPIDTO(BaseModel):
    id: str = Field(alias="payOrderId")
    amount: AmountInfo = Field(alias="amount")
    amount_transferred: AmountInfo = Field(alias="amountTransferred")
    client: ClientInfo = Field(alias="client")
    created_at: str = Field(alias="createdAt")
    debt_id: str = Field(alias="debtId")
    disbursement_date: str = Field(alias="disbursementDate")
    gateway: str = Field(alias="gateway")
    notes: Optional[str] = Field(alias="notes", default=None)
    partner: PartnerInfo = Field(alias="partner")
    payer: Optional[PayerInfo] = Field(alias="payer", default=None)
    beneficiary: Optional[PayerInfo] = Field(alias="beneficiary", default=None)
    payment_date: Optional[str] = Field(alias="paymentDate", default=None)
    reference_id: str = Field(alias="referenceId")
    reference_number: int = Field(alias="referenceNumber")
    status: str = Field(alias="status")
    tracking_key: str = Field(alias="trackingKey")
    updated_at: str = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class DisbursementBase:
    @staticmethod
    def _conciliate_disbursement(payOrderId: str):
        return f"/v1/disbursements/{payOrderId}/reconcile"
    
    @staticmethod
    def _disbursement_success(payOrderId: str):
        return f"/v1/disbursements/{payOrderId}/success"

class DisbursementAsync(DisbursementBase):
    data: DisbursementAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DisbursementAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"
    
    @retry_on_401_async
    async def conciliate_disbursement(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.put(
                self._conciliate_disbursement(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def disbursement_success(self):
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.put(
                self._disbursement_success(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)

class DisbursementSync(DisbursementBase):
    data: DisbursementAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DisbursementAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"
    
    @retry_on_401
    def conciliate_disbursement(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.put(
                self._conciliate_disbursement(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
    
    @retry_on_401
    def disbursement_success(self):
        with httpx.Client(base_url=self.base_url) as client:
            response = client.put(
                self._disbursement_success(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)

class DisbursementAsyncModule(GenericAsyncModule):
    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=DisbursementAsync,
            retrieve_data_model=DisbursementAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="disbursements"
        )

class DisbursementSyncModule(GenericSyncModule):
    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=DisbursementSync,
            retrieve_data_model=DisbursementAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="disbursements"
        )


from typing import List, Optional
from pydantic import BaseModel, Field


# ---------------------------------------------------------------------------
#  Detailed split-bucket money model
#  • Snake-case attribute names for Pythonic access.
#  • camelCase JSON aliases so front-end / API payloads remain idiomatic.
#  • Every numeric value is a STRING to keep original precision and formatting.
# ---------------------------------------------------------------------------


class TaxBreakdown(BaseModel):
    """
    One row per individual tax, withholding, duty, levy, etc.
    """

    # --- identifiers -------------------------------------------------------
    label: str = Field(alias="label", description="Human-readable label")
    code: Optional[str] = Field(default=None, alias="code", description="Short tax code, e.g. 'VAT21'")
    category: Optional[str] = Field(
        default=None,
        alias="category",
        description="Classification bucket: VAT | TURNOVER | EXCISE | WH (withholding) …",
    )
    jurisdiction: Optional[str] = Field(
        default=None,
        alias="jurisdiction",
        description="ISO-3166-2 or custom region code where the tax applies",
    )

    # --- monetary & rate ---------------------------------------------------
    rate: Optional[str] = Field(
        default="0",
        alias="rate",
        description="Rate as string (e.g. '0.21' for 21 %) – string keeps exact decimals",
    )
    base: Optional[str] = Field(
        default="0",
        alias="base",
        description="Taxable base amount (string)",
    )
    amount: str = Field(
        default="0",
        alias="amount",
        description="Tax amount (string) – NEGATIVE for withholdings",
    )

    # --- optional per-row FX ----------------------------------------------
    currency: Optional[str] = Field(
        default=None,
        alias="currency",
        description="Currency for this row if different from document header",
    )
    fx_rate: Optional[str] = Field(
        default=None,
        alias="fxRate",
        description="FX rate applied if row currency ≠ functional currency",
    )

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class Amount(BaseModel):
    """
    Split-bucket monetary object that works for ANY document
    (invoice, credit note, payment order, etc.).

    Math guard-rails:
      gross  = net + tax
      payable = gross − discount − withholding − prepaid
      tax     = sum(tax_breakdown.amount)

    Store values as strings to avoid rounding drift; cast to Decimal as needed.
    """

    # --- always present ----------------------------------------------------
    net: Optional[str] = Field(
        default="0",
        alias="net",
        description="Base amount before taxes or deductions",
    )
    gross: Optional[str] = Field(
        default="0",
        alias="gross",
        description="net + tax (before discounts / withholdings)",
    )
    currency: str = Field(
        ...,
        alias="currency",
        description="ISO-4217 currency code, e.g. 'ARS'",
    )

    # --- often zero, so default "0" ---------------------------------------
    tax: Optional[str] = Field(
        default="0",
        alias="tax",
        description="Total taxes (positive) or sum of withholdings (negative rows included)",
    )
    discount: Optional[str] = Field(
        default="0",
        alias="discount",
        description="Header-level commercial discounts / allowances",
    )
    withholding: Optional[str] = Field(
        default="0",
        alias="withholding",
        description="Total retained taxes that reduce cash out (positive string even though accounting sign is negative)",
    )
    prepaid: Optional[str] = Field(
        default="0",
        alias="prepaid",
        description="Deposits or advances already applied to this document",
    )
    payable: Optional[str] = Field(
        default="0",
        alias="payable",
        description="Cash to be paid now = gross − discount − withholding − prepaid",
    )

    # --- functional-currency helpers --------------------------------------
    fx_rate: Optional[str] = Field(
        default="0",
        alias="fxRate",
        description="Exchange rate to functional currency; '0' if not applicable",
    )

    # --- granular rows -----------------------------------------------------
    tax_breakdown: List[TaxBreakdown] = Field(
        default_factory=list,
        alias="taxBreakdown",
        description="Detailed list of each tax / withholding with base and amount",
    )

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True



from pydantic import BaseModel, Field
from typing import Optional, Dict, Any


class AttachmentAPIDTO(BaseModel):
    id: str = Field(alias="id")
    url: Optional[str] = Field(alias="url", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    file_extension: Optional[str] = Field(alias="fileExtension", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata")
    created_at: str = Field(alias="createdAt", default=str)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class AttachmentInput(BaseModel):
    url: str = Field(alias="url")
    label: Optional[str] = Field(alias="label", default=None)
    file_extension: Optional[str] = Field(alias="fileExtension")
    metadata: Optional[dict] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True



from pydantic import BaseModel, Field
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from typing import Optional, Dict
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async


class AddressAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    label: Optional[str] = Field(alias="label")
    street1: Optional[str] = Field(alias="street1", default=None)
    street2: Optional[str] = Field(alias="street2", default=None)
    neighborhood: Optional[str] = Field(alias="neighborhood", default=None)
    city: Optional[str] = Field(alias="city", default=None)
    state: Optional[str] = Field(alias="state", default=None)
    zip_code: Optional[str] = Field(alias="zipCode", default=None)
    reference: Optional[str] = Field(alias="reference", default=None)
    country: Optional[str] = Field(alias="country", default=None)
    province: Optional[str] = Field(alias="province", default=None)
    lat: Optional[float] = Field(alias="lat", default=None)
    lon: Optional[float] = Field(alias="lon", default=None)
    priority: Optional[int] = Field(alias="priority", default=None)
    is_home: bool = Field(alias="isHome", default=False)
    is_work: bool = Field(alias="isWork", default=False)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")
    has_attachments: bool = Field(alias="hasAttachments")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    def get_location(self):
        if self.lat is None or self.lon is None:
            return None
        return {
            "lat": self.lat,
            "lon": self.lon
        }

    def get_address(self):
        return {
            "street1": self.street1,
            "street2": self.street2,
            "neighborhood": self.neighborhood,
            "city": self.city,
            "province": self.province,
            "state": self.state,
            "zipCode": self.zip_code,
            "country": self.country,
        }

    def get_address_str(self):
        str = ""
        if self.street1 is not None:
            str += self.street1
        if self.street2 is not None:
            str += " " + self.street2
        if self.neighborhood is not None:
            str += " " + self.neighborhood
        if self.city is not None:
            str += " " + self.city
        if self.province is not None:
            str += " " + self.province
        if self.state is not None:
            str += " " + self.state
        if self.zip_code is not None:
            str += " " + self.zip_code
        if self.country is not None:
            str += " " + self.country
        return str


class CreateAddressDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    label: Optional[str] = Field(alias="label")
    street1: Optional[str] = Field(alias="street1", default=None)
    street2: Optional[str] = Field(alias="street2", default=None)
    neighborhood: Optional[str] = Field(alias="neighborhood", default=None)
    city: Optional[str] = Field(alias="city", default=None)
    state: Optional[str] = Field(alias="state", default=None)
    zip_code: Optional[str] = Field(alias="zipCode", default=None)
    reference: Optional[str] = Field(alias="reference", default=None)
    country: Optional[str] = Field(alias="country", default=None)
    province: Optional[str] = Field(alias="province", default=None)
    lat: Optional[float] = Field(alias="lat", default=None)
    lon: Optional[float] = Field(alias="lon", default=None)
    priority: Optional[int] = Field(alias="priority", default=None)
    is_home: bool = Field(alias="isHome", default=False)
    is_work: bool = Field(alias="isWork", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateAddressDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    label: Optional[str] = Field(alias="label", default=None)
    street1: Optional[str] = Field(alias="street1", default=None)
    street2: Optional[str] = Field(alias="street2", default=None)
    neighborhood: Optional[str] = Field(alias="neighborhood", default=None)
    city: Optional[str] = Field(alias="city", default=None)
    state: Optional[str] = Field(alias="state", default=None)
    zip_code: Optional[str] = Field(alias="zipCode", default=None)
    reference: Optional[str] = Field(alias="reference", default=None)
    country: Optional[str] = Field(alias="country", default=None)
    lat: Optional[float] = Field(alias="lat", default=None)
    lon: Optional[float] = Field(alias="lon", default=None)
    priority: Optional[int] = Field(alias="priority", default=None)
    is_home: Optional[bool] = Field(alias="isHome", default=None)
    is_work: Optional[bool] = Field(alias="isWork", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class GeocodingAPIDTO(BaseModel):
    id: str = Field(alias="id")
    not_found: Optional[bool] = Field(alias="notFound", default=None)
    is_success: bool = Field(alias="isSuccess")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class AddressSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "addresses", header_builder, renew_token, AddressAPIDTO.parse_obj(data))


class AddressAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "addresses", header_builder, renew_token, AddressAPIDTO.parse_obj(data))


class AddressesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=AddressSync,
                         retrieve_data_model=AddressAPIDTO,
                         create_data_model=CreateAddressDTO,
                         update_data_model=UpdateAddressDTO,
                         resource="addresses")


    @retry_on_401
    def geocode(self, address_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                f"/v1/addresses/commands/geocode",
                headers=self.build_headers(),
                json={
                    "id": address_id
                }
            )
            raise_for_status_improved(response)
            return GeocodingAPIDTO.parse_obj(response.json())

    @retry_on_401
    def reverse_geocode(self, address_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.post(
                f"/v1/addresses/commands/reverse-geocode",
                headers=self.build_headers(),
                json={
                    "id": address_id
                }
            )
            raise_for_status_improved(response)
            return GeocodingAPIDTO.parse_obj(response.json())


class AddressesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=AddressAsync,
                         retrieve_data_model=AddressAPIDTO,
                         create_data_model=CreateAddressDTO,
                         update_data_model=UpdateAddressDTO,
                         resource="addresses")


    @retry_on_401_async
    async def geocode(self, address_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/addresses/commands/geocode",
                headers=self.build_headers(),
                json={
                    "id": address_id
                }
            )
            raise_for_status_improved(response)
            return GeocodingAPIDTO.parse_obj(response.json())

    @retry_on_401_async
    async def reverse_geocode(self, address_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.post(
                f"/v1/addresses/commands/reverse-geocode",
                headers=self.build_headers(),
                json={
                    "id": address_id
                }
            )
            raise_for_status_improved(response)
            return GeocodingAPIDTO.parse_obj(response.json())
